@@ -19,7 +19,9 @@
 #include "aff_data.hxx"
 #include "utils.hxx"
 
+#include <charconv>
 #include <iostream>
+#include <locale>
 #include <sstream>
 #include <unordered_map>
 
@@ -70,7 +72,7 @@ enum class Parsing_Error_Code {
 	COMPOUND_RULE_INVALID_FORMAT
 };
 
-auto decode_flags(const string& s, Flag_Type t, const Encoding& enc,
+auto decode_flags(string_view s, Flag_Type t, const Encoding& enc,
                   u16string& out) -> Parsing_Error_Code
 {
 	using Err = Parsing_Error_Code;
@@ -108,30 +110,22 @@ auto decode_flags(const string& s, Flag_Type t, const Encoding& enc,
 		}
 		break;
 	}
-	case Ft::NUMBER: {
-		auto p = s.c_str();
-		char* p2 = nullptr;
-		errno = 0;
-		for (;;) {
-			auto flag = strtoul(p, &p2, 10);
-			if (p2 == p)
+	case Ft::NUMBER:
+		for (auto p = begin_ptr(s);;) {
+			uint16_t flag;
+			auto fc = from_chars(p, end_ptr(s), flag);
+			if (fc.ec == errc::invalid_argument)
 				return Err::INVALID_NUMERIC_FLAG;
-			if (flag == numeric_limits<decltype(flag)>::max() &&
-			    errno == ERANGE) {
-				errno = 0;
-				return Err::FLAG_ABOVE_65535;
-			}
-			if (flag > 0xFFFF)
+			if (fc.ec == errc::result_out_of_range)
 				return Err::FLAG_ABOVE_65535;
 			out.push_back(flag);
 
-			if (p2 == end_ptr(s) || *p2 != ',')
+			if (fc.ptr == end_ptr(s) || *fc.ptr != ',')
 				break;
 
-			p = p2 + 1;
+			p = fc.ptr + 1;
 		}
 		break;
-	}
 	case Ft::UTF8: {
 		// if (!enc.is_utf8())
 		//	return Err::FLAGS_ARE_UTF8_BUT_FILE_NOT;
@@ -152,7 +146,7 @@ auto decode_flags(const string& s, Flag_Type t, const Encoding& enc,
 	return warn;
 }
 
-auto decode_flags_possible_alias(const string& s, Flag_Type t,
+auto decode_flags_possible_alias(string_view s, Flag_Type t,
                                  const Encoding& enc,
                                  const vector<Flag_Set>& flag_aliases,
                                  u16string& out) -> Parsing_Error_Code
@@ -160,14 +154,11 @@ auto decode_flags_possible_alias(const string& s, Flag_Type t,
 	if (flag_aliases.empty())
 		return decode_flags(s, t, enc, out);
 
-	char* p;
-	errno = 0;
 	out.clear();
-	auto i = strtoul(s.c_str(), &p, 10);
-	if (p == s.c_str())
-		return Parsing_Error_Code::INVALID_NUMERIC_ALIAS;
-
-	if (i == numeric_limits<decltype(i)>::max() && errno == ERANGE)
+	size_t i;
+	auto fc = from_chars(begin_ptr(s), end_ptr(s), i);
+	if (fc.ec == errc::invalid_argument ||
+	    fc.ec == errc::result_out_of_range)
 		return Parsing_Error_Code::INVALID_NUMERIC_ALIAS;
 
 	if (0 < i && i <= flag_aliases.size()) {
@@ -237,7 +228,7 @@ auto report_parsing_error(Parsing_Error_Code err, size_t line_num)
 	}
 }
 
-auto decode_compound_rule(const string& s, Flag_Type t, const Encoding& enc,
+auto decode_compound_rule(string_view s, Flag_Type t, const Encoding& enc,
                           u16string& out) -> Parsing_Error_Code
 {
 	using Ft = Flag_Type;
@@ -272,27 +263,23 @@ auto decode_compound_rule(const string& s, Flag_Type t, const Encoding& enc,
 		out.clear();
 		if (s.empty())
 			return Err::MISSING_FLAGS;
-		errno = 0;
-		for (auto p = s.c_str(); *p != 0;) {
+		for (auto p = begin_ptr(s); p != end_ptr(s);) {
 			if (*p != '(')
 				return Err::COMPOUND_RULE_INVALID_FORMAT;
 			++p;
-			char* p2;
-			auto flag = strtoul(p, &p2, 10);
-			if (p2 == p)
+			uint16_t flag;
+			auto fc = from_chars(p, end_ptr(s), flag);
+			if (fc.ec == errc::invalid_argument)
 				return Err::INVALID_NUMERIC_FLAG;
-			if (flag == numeric_limits<decltype(flag)>::max() &&
-			    errno == ERANGE) {
-				errno = 0;
+			if (fc.ec == errc::result_out_of_range)
 				return Err::FLAG_ABOVE_65535;
-			}
-			if (flag > 0xFFFF)
-				return Err::FLAG_ABOVE_65535;
-			p = p2;
-			if (*p != ')')
+			p = fc.ptr;
+			if (p == end_ptr(s) || *p != ')')
 				return Err::COMPOUND_RULE_INVALID_FORMAT;
 			out.push_back(flag);
 			++p;
+			if (p == end_ptr(s))
+				break;
 			if (*p == '?' || *p == '*') {
 				out.push_back(*p);
 				++p;
@@ -778,7 +765,6 @@ auto Aff_Data::parse_aff(istream& in) -> bool
 	auto line_num = size_t(0);
 	auto ss = istringstream();
 	auto p = Aff_Line_IO_Manip(*this);
-	Setlocale_To_C_In_Scope setlocale_to_C;
 	auto error_happened = false;
 	// while parsing, the streams must have plain ascii locale without
 	// any special number separator otherwise istream >> int might fail
@@ -928,37 +914,6 @@ auto Aff_Data::parse_aff(istream& in) -> bool
 	return in.eof() && !error_happened; // true for success
 }
 
-/**
- * @internal
- * @brief Scans @p line for morphological field [a-z][a-z]:
- *
- * Scans the line for space, two lowercase alphabetic chars and colon.
- *
- * @param line
- * @returns the end of the word before the morph field, or npos
- */
-auto dic_find_end_of_word_heuristics(const string& line)
-{
-	if (line.size() < 4)
-		return line.npos;
-	size_t a = 0;
-	for (;;) {
-		a = line.find(' ', a);
-		if (a == line.npos)
-			break;
-		auto b = line.find_first_not_of(' ', a);
-		if (b == line.npos)
-			break;
-		if (b > line.size() - 3)
-			break;
-		if (line[b] >= 'a' && line[b] <= 'z' && line[b + 1] >= 'a' &&
-		    line[b + 1] <= 'z' && line[b + 2] == ':')
-			return a;
-		a = b;
-	}
-	return line.npos;
-}
-
 auto Aff_Data::parse_dic(istream& in) -> bool
 {
 	size_t line_number = 1;
@@ -969,11 +924,11 @@ auto Aff_Data::parse_dic(istream& in) -> bool
 	u16string flags;
 	string u8word;
 	auto enc_conv = Encoding_Converter(encoding.value_or_default());
+	auto success = true;
 
 	// locale must be without thousands separator.
 	auto& ctype = use_facet<std::ctype<char>>(locale::classic());
 	in.imbue(locale::classic());
-	Setlocale_To_C_In_Scope setlocale_to_C;
 
 	strip_utf8_bom(in);
 	if (in >> approximate_size)
@@ -987,25 +942,46 @@ auto Aff_Data::parse_dic(istream& in) -> bool
 		word.clear();
 		flags_str.clear();
 		flags.clear();
-		if (!line.empty() && line.back() == '\r')
+		if (!empty(line) && line.back() == '\r')
 			line.pop_back();
 
-		size_t slash_pos = 0;
-		size_t tab_pos = 0;
-		for (;;) {
-			slash_pos = line.find('/', slash_pos);
-			if (slash_pos == line.npos)
+		auto end_word_pos = line.npos;
+		for (size_t i = 0; i != size(line); ++i) {
+			switch (line[i]) {
+			case '/':
+				if (i == 0)
+					continue;
+				if (line[i - 1] == '\\') {
+					--i;
+					line.erase(i, 1);
+				}
+				else {
+					end_word_pos = i;
+				}
 				break;
-			if (slash_pos == 0)
+			case '\t':
+				end_word_pos = i;
 				break;
-			if (line[slash_pos - 1] != '\\')
+			case ' ': {
+				auto p = ctype.scan_not(
+				    ctype.space, &line[i + 1], end_ptr(line));
+				size_t k = p - begin_ptr(line);
+				if (k == size(line) ||
+				    (size(line) - k >= 3 &&
+				     line[k + 2] == ':' &&
+				     ctype.is(ctype.lower, line[k]) &&
+				     ctype.is(ctype.lower, line[k + 1])))
+					end_word_pos = i;
+				break;
+			}
+			}
+			if (end_word_pos != line.npos)
 				break;
-
-			line.erase(slash_pos - 1, 1);
 		}
-		if (slash_pos != line.npos && slash_pos != 0) {
+		word.assign(line, 0, end_word_pos);
+		if (end_word_pos != line.npos && line[end_word_pos] == '/') {
 			// slash found, word until slash
-			word.assign(line, 0, slash_pos);
+			auto slash_pos = end_word_pos;
 			auto ptr = ctype.scan_is(ctype.space, &line[slash_pos],
 			                         end_ptr(line));
 			auto end_flags_pos = ptr - begin_ptr(line);
@@ -1014,20 +990,21 @@ auto Aff_Data::parse_dic(istream& in) -> bool
 			auto err = decode_flags_possible_alias(
 			    flags_str, flag_type, encoding, flag_aliases,
 			    flags);
+			if (err == Parsing_Error_Code::MISSING_FLAGS ||
+
+			    (/* bug in eu.dic that we fix here. Remove this once
+			        fixed there. */
+			     err == Parsing_Error_Code::INVALID_NUMERIC_FLAG &&
+			     flags_str == "None"))
+				err = Parsing_Error_Code::
+				    NO_FLAGS_AFTER_SLASH_WARNING;
 			report_parsing_error(err, line_number);
-			if (static_cast<int>(err) > 0)
+			if (static_cast<int>(err) > 0) {
+				success = false;
 				continue;
+			}
 		}
-		else if ((tab_pos = line.find('\t')) != line.npos) {
-			// Tab found, word until tab. No flags.
-			// After tab follow morphological fields
-			word.assign(line, 0, tab_pos);
-		}
-		else {
-			auto end = dic_find_end_of_word_heuristics(line);
-			word.assign(line, 0, end);
-		}
-		if (word.empty())
+		if (empty(word))
 			continue;
 		auto ok = enc_conv.to_utf8(word, u8word);
 		if (!ok)
@@ -1058,7 +1035,7 @@ auto Aff_Data::parse_dic(istream& in) -> bool
 			break;
 		}
 	}
-	return in.eof(); // success if we reached eof
+	return in.eof() && success; // success if we reached eof
 }
 } // namespace v5
 } // namespace nuspell
